"
 COPYRIGHT (c) 1988 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

LimitedPrecisionReal variableByteSubclass:#Float
       instanceVariableNames:''
       classVariableNames:''
       poolDictionaries:''
       category:'Magnitude-Numbers'
!

Float comment:'

COPYRIGHT (c) 1988 by Claus Gittinger
              All Rights Reserved

$Header: /cvs/stx/stx/libbasic/Float.st,v 1.11 1994-02-25 12:58:48 claus Exp $

notice, that Floats are defined as Byte-array to prevent the garbage collector
from going into the value ... otherwise I needed a special case in many places.
Also notice, that ST/X Floats are what Doubles are in ST-80 - this may change 
in one of the next versions.
'!

!Float class methodsFor:'instance creation'!

basicNew
    "return a new float - here we return 0.0
     - floats are usually NOT created this way ...
     Its implemented here to allow things like binary store & load
     of floats. (but even this support will go away eventually, its not
     a good idea to store the bits of a float - the reader might have a
     totally different representation - so floats will eventually be 
     binary stored in a device independent format."

%{  /* NOCONTEXT */
    RETURN ( _MKFLOAT((0.0) COMMA_SND) );
%}
! !

!Float class methodsFor:'constants'!

zero
    "return the neutral element for addition"

    ^ 0.0
!

unity
    "return the neutral element for multiplication"

    ^ 1.0
!

pi
    "return the constant pi"

    ^ 3.1415926535897932384626434
! !

!Float class methodsFor:'queries'!

isBuiltInClass
    "this class is known by the run-time-system"

    ^ self == Float
! !

!Float methodsFor:'arithmetic'!

+ aNumber
    "return the sum of the receiver and the argument, aNumber"

%{  /* NOCONTEXT */

    OBJ newFloat;
    double result;

    if (_isSmallInteger(aNumber)) {
        result = _floatVal(self) + (double)(_intVal(aNumber));
retResult:
        _qMKFLOAT(newFloat, result, SENDER);
        RETURN ( newFloat );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        result = _floatVal(self) + _floatVal(aNumber);
        goto retResult;
    }
%}
.
    ^ aNumber sumFromFloat:self
!

- aNumber
    "return the difference of the receiver and the argument, aNumber"

%{  /* NOCONTEXT */

    OBJ newFloat;
    double result;

    if (_isSmallInteger(aNumber)) {
        result = _floatVal(self) - (double)(_intVal(aNumber));
retResult:
        _qMKFLOAT(newFloat, result, SENDER);
        RETURN ( newFloat );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        result = _floatVal(self) - _floatVal(aNumber);
        goto retResult;
    }
%}
.
    ^ aNumber differenceFromFloat:self
!

* aNumber
    "return the product of the receiver and the argument, aNumber"

%{  /* NOCONTEXT */

    OBJ newFloat;
    double result;

    if (_isSmallInteger(aNumber)) {
        result = _floatVal(self) * (double)(_intVal(aNumber));
retResult:
        _qMKFLOAT(newFloat, result, SENDER);
        RETURN ( newFloat );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        result = _floatVal(self) * _floatVal(aNumber);
        goto retResult;
    }
%}
.
    ^ aNumber productFromFloat:self
!

/ aNumber
    "return the quotient of the receiver and the argument, aNumber"

%{  /* NOCONTEXT */

    OBJ newFloat;
    double result, val;

    if (_isSmallInteger(aNumber)) {
        if (aNumber != _MKSMALLINT(0)) {
            result = _floatVal(self) / ( (double)_intVal(aNumber)) ;
retResult:
            _qMKFLOAT(newFloat, result, SENDER);
            RETURN ( newFloat );
        }
    } else {
        if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
            val = _floatVal(aNumber);
            if (val != 0.0) {
                result = _floatVal(self) / val;
                goto retResult;
            }
        }
    }
%}
.
    ((aNumber == 0) or:[aNumber = 0.0]) ifTrue:[
        DivisionByZeroSignal raise.
        ^ self
    ].
    ^ aNumber quotientFromFloat:self
!

// aNumber
    "return the integer quotient of dividing the receiver by aNumber with
    truncation towards negative infinity."

    ^ (self / aNumber) floor asInteger
!

\\ aNumber
    "return the integer remainder of dividing the receiver by aNumber with
    truncation towards negative infinity."

    ^ (self - ((self / aNumber) floor * aNumber)) floor asInteger
!

negated
    "return myself negated"

%{  /* NOCONTEXT */

    OBJ newFloat;

    _qMKFLOAT(newFloat, - _floatVal(self), SENDER);
    RETURN ( newFloat );
%}
! !

!Float methodsFor:'testing'!

positive
    "return true if the receiver is greater or equal to zero"

%{  /* NOCONTEXT */

    RETURN ( (_floatVal(self) >= 0.0) ? true : false );
%}
!

negative
    "return true if the receiver is less than zero"

%{  /* NOCONTEXT */

    RETURN ( (_floatVal(self) < 0.0) ? true : false );
%}
! !

!Float methodsFor:'comparing'!

< aNumber
    "return true, if the argument is greater"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aNumber)) {
        RETURN ( (_floatVal(self) < (double)(_intVal(aNumber))) ? true : false );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        RETURN ( (_floatVal(self) < _floatVal(aNumber)) ? true : false );
    }
%}
.
    ^ aNumber lessFromFloat:self
!

> aNumber
    "return true, if the argument is less"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aNumber)) {
        RETURN ( (_floatVal(self) > (double)(_intVal(aNumber))) ? true : false );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        RETURN ( (_floatVal(self) > _floatVal(aNumber)) ? true : false );
    }
%}
.
    ^ self retry:#> coercing:aNumber
!

<= aNumber
    "return true, if the argument is greater or equal"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aNumber)) {
        RETURN ( (_floatVal(self) <= (double)(_intVal(aNumber))) ? true : false );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        RETURN ( (_floatVal(self) <= _floatVal(aNumber)) ? true : false );
    }
%}
.
    ^ self retry:#<= coercing:aNumber
!

>= aNumber
    "return true, if the argument is less or equal"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aNumber)) {
        RETURN ( (_floatVal(self) >= (double)(_intVal(aNumber))) ? true : false );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        RETURN ( (_floatVal(self) >= _floatVal(aNumber)) ? true : false );
    }
%}
.
    ^ self retry:#>= coercing:aNumber
!

= aNumber
    "return true, if the arguments value are equal"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aNumber)) {
        RETURN ( (_floatVal(self) == (double)(_intVal(aNumber))) ? true : false );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        RETURN ( (_floatVal(self) == _floatVal(aNumber)) ? true : false );
    }
%}
.
    ^ self retry:#= coercing:aNumber
!

~= aNumber
    "return true, if the arguments value are not equal"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aNumber)) {
        RETURN ( (_floatVal(self) != (double)(_intVal(aNumber))) ? true : false );
    }
    if ((aNumber != nil) && (_qClass(aNumber) == Float)) {
        RETURN ( (_floatVal(self) != _floatVal(aNumber)) ? true : false );
    }
%}
.
    ^ self retry:#~= coercing:aNumber
! !

!Float methodsFor:'coercion and converting'!

coerce:aNumber
    "return aNumber converted into receivers type"

    ^ aNumber asFloat
!

generality
    "return the generality value - see ArithmeticValue>>retry:coercing:"

    ^ 80
!

asFloat
    "return a float with same value - thats me"

    ^ self
!

asInteger
    "return an integer with same value - might truncate"

    |l v sign|

%{  /* NOCONTEXT */

    if ((_floatVal(self) >= (double)_MIN_INT)
     && (_floatVal(self) <= (double)_MAX_INT)) {
        RETURN ( _MKSMALLINT( (INT)_floatVal(self)) );
    }
%}
.
    ^ super asInteger

    "12345.0 asInteger"
    "1e15 asInteger"
! !

!Float methodsFor:'truncation and rounding'!

truncated
    "return the receiver truncated towards zero as Integer"

    |val|

%{
    double floor(), ceil();
    double dVal;

    dVal = _floatVal(self);
    if (dVal < 0.0) {
        dVal = ceil(dVal);
    } else {
        dVal = floor(dVal);
    }

    /*
     * mhmh it seems that ST-80 is returning integers if possible.
     * (at least, some pd programs expect it ...)
     */
    if ((dVal >= (double)_MIN_INT) && (dVal <= (double)_MAX_INT)) {
        RETURN ( _MKSMALLINT( (INT) dVal ) );
    }
    val = _MKFLOAT(dVal COMMA_CON);
%}
.
    ^ val asInteger
!

rounded
    "return the receiver rounded to the nearest integer as integer"

    |val|

%{
    double floor();
    double dVal;

    /*
     * mhmh it seems that ST-80 is returning integers if possible
     * at least, some pd programs expect it ...
     */
    dVal = floor(_floatVal(self) + 0.5);
    if ((dVal >= (double)_MIN_INT) && (dVal <= (double)_MAX_INT)) {
        RETURN ( _MKSMALLINT( (INT) dVal ) );
    }
    val = _MKFLOAT(dVal COMMA_CON);
%}
.
    ^ val asInteger
!

floor
    "return the biggest integer-valued float less or equal to the receiver"

%{  /* NOCONTEXT */

    double floor();

    RETURN ( _MKFLOAT(floor(_floatVal(self)) COMMA_SND) );
%}
!

ceiling
    "return the smallest integer-valued float greater or equal to the receiver"

%{  /* NOCONTEXT */

    double ceil();

    RETURN ( _MKFLOAT(ceil(_floatVal(self)) COMMA_SND) );
%}
! !

!Float methodsFor:'mathematical functions'!

ln
    "return the natural logarithm of myself"

%{  /* NOCONTEXT */

    double log();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = log(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
!

raisedTo:aNumber
    "return self raised to the power of aNumber"
    |n|

    n := aNumber asFloat.
%{
    double pow();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    if (__isFloat(n)) {
        errno = 0;
        result = pow(_floatVal(self), _floatVal(n));
        errno = 0;  /* XXXX */
        if (errno == 0)
            RETURN ( _MKFLOAT(result COMMA_CON) );
        ErrorNumber = _MKSMALLINT(errno);
    }
%}
.
    DomainErrorSignal raise
!

exp
    "return e raised to the power of the receiver"

%{  /* NOCONTEXT */

    double exp();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = exp(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
!

sin
    "return the sine of myself interpreted as radians"

%{  /* NOCONTEXT */

    double sin();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = sin(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
!

cos
    "return the cosine of myself interpreted as radians"

%{  /* NOCONTEXT */

    double cos();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = cos(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
!

tan
    "return the tangent of myself interpreted as radians"

%{  /* NOCONTEXT */

    double tan();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = tan(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
!

arcSin
    "return the arcsine of myself as radians"

%{  /* NOCONTEXT */

    double asin();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = asin(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
!

arcCos
    "return the arccosine of myself as radians"

%{  /* NOCONTEXT */

    double acos();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = acos(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
!

arcTan
    "return the arctangent of myself as radians"

%{  /* NOCONTEXT */

    double atan();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = atan(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
!

sqrt
    "return the square root of myself"

%{  /* NOCONTEXT */

    double sqrt();
    double result;
    extern errno;
    extern OBJ ErrorNumber;

    errno = 0;
    result = sqrt(_floatVal(self));
    if (errno == 0)
        RETURN ( _MKFLOAT(result COMMA_SND) );
    ErrorNumber = _MKSMALLINT(errno);
%}
.
    DomainErrorSignal raise
! !

!Float class methodsFor:'binary storage'!

storeBinaryIEEESingle:aFloat on:aStream
    "store aFloat as an IEEE formatted 4-byte float
     onto the binary stream, aStream"

    "this implementation is wrong: dous not work on non-IEEE machines"
    UninterpretedBytes isBigEndian ifTrue:[
        "swap the bytes"
        8 to:4 by:-1 do:[:i |
            aStream nextPut:(aFloat basicAt:i).
        ].
        ^ self
    ].
    4 to:8 do:[:i |
        aStream nextPut:(aFloat basicAt:i).
    ]
!

storeBinaryIEEEDouble:aFloat on:aStream
    "store aFloat as an IEEE formatted 8-byte float
     onto the binary stream, aStream"

    "this implementation is wrong: dous not work on non-IEEE machines"
    UninterpretedBytes isBigEndian ifTrue:[
        "swap the bytes"
        8 to:1 by:-1 do:[:i |
            aStream nextPut:(aFloat basicAt:i).
        ].
        ^ self
    ].
    1 to:8 do:[:i |
        aStream nextPut:(aFloat basicAt:i).
    ].
!

readBinaryIEEESingleFrom:aStream into:aFloat
    "read a float value from the binary stream, aStream,
     interpreting the next bytes as an IEEE formatted 4-byte float"

    UninterpretedBytes isBigEndian ifTrue:[
        "swap the bytes"
        8 to:4 by:-1 do:[:i |
            aFloat basicAt:i put:(aStream next)
        ].
        ^ self
    ].
    4 to:8 do:[:i |
        aFloat basicAt:i put:aStream next
    ]
!

readBinaryIEEEDoubleFrom:aStream into:aFloat
    "read the receivers value from the binary stream, aStream,
     interpreting the next bytes as an IEEE formatted 8-byte float"

    UninterpretedBytes isBigEndian ifTrue:[
        "swap the bytes"
        8 to:1 by:-1 do:[:i |
            aFloat basicAt:i put:(aStream next)
        ].
        ^ self
    ].
    1 to:8 do:[:i |
        aFloat basicAt:i put:aStream next
    ]
!

binaryDefinitionFrom: stream manager: manager
    |f|

    f := self basicNew.
    self readBinaryIEEEDoubleFrom:stream into:f.
    ^ f
! !

!Float methodsFor:'binary storage'!

storeBinaryDefinitionOn: stream manager: manager
    manager putIdOf: self class on: stream.
    Float storeBinaryIEEEDouble:self on:stream.
! !

!Float methodsFor:'printing and storing'!

printString
    "return a printed representation of the receiver"

%{  /* NOCONTEXT */

    char buffer[64];
    REGISTER char *cp;
#ifdef THISCONTEXT_IN_REGISTER
    /*
     * actually only needed on sparc: since thisContext is
     * in a global register, which gets destroyed by printf,
     * manually save it here - very stupid ...
     */
    OBJ sav = __thisContext;
#endif

#ifdef SYSV
    sprintf(buffer, "%.6lg", _floatVal(self));
#else
    sprintf(buffer, "%.6G", _floatVal(self));
#endif

#ifdef THISCONTEXT_IN_REGISTER
    __thisContext = sav;
#endif
    /* 
     * kludge to make integral float f prints as "f.0" (not as "f" as printf does)
     * (i.e. look if string contains '.' or 'e' and append '.0' if not)
     */
    for (cp = buffer; *cp; cp++) {
        if ((*cp == '.') || (*cp == 'e')) break;
    }
    if (! *cp) {
        *cp++ = '.';
        *cp++ = '0';
        *cp = '\0';
    }

    RETURN ( _MKSTRING(buffer COMMA_SND) );
%}
    "1.0 printString"
    "1.234 printString"
    "1e10 printString"
    "1.2e3 printString"
    "1.2e30 printString"
!

printfPrintString:formatString
    "non-portable: return a printed representation of the receiver
     as specified by formatString, which is defined by printf.
     If you use this, be aware, that specifying doubles differs on
     systems; on SYSV machines you have to give something like %lf, 
     while on BSD systems the format string has to be %F.
     Also, the resulting string may not be longer than 255 bytes -
     since thats the (static) size of the buffer.
     This method is NONSTANDARD and may be removed without notice."

%{  /* STACK: 400 */
    char buffer[256];

    if (__isString(formatString)) {
        /*
         * actually only needed on sparc: since thisContext is
         * in a global register, which gets destroyed by printf,
         * manually save it here - very stupid ...
         */
        OBJ sav = __thisContext;

        sprintf(buffer, _stringVal(formatString), _floatVal(self));
        __thisContext = sav;
        RETURN ( _MKSTRING(buffer COMMA_SND) );
    }
%}
.
    self primitiveFailed

    "Float pi printfPrintString:'%%lg -> %lg'"
    "Float pi printfPrintString:'%%lf -> %lf'"
    "Float pi printfPrintString:'%%7.5lg -> %7.5lg'"
    "Float pi printfPrintString:'%%7.5lf -> %7.5lf'"
    "Float pi printfPrintString:'%%G -> %G'"
    "Float pi printfPrintString:'%%F -> %F'"
    "Float pi printfPrintString:'%%7.5G -> %7.5G'"
    "Float pi printfPrintString:'%%7.5F -> %7.5F'"
! !
